"""airplane - An SDK for writing Airplane tasks in Python"""

from airplane import auth, display, files
from airplane._version import __version__
from airplane.api.client import APIClient
from airplane.api.entities import PromptReviewers, Run, RunStatus
from airplane.builtins import email, graphql, mongodb, rest, slack, sql
from airplane.config import EnvVar, Resource, Schedule, task
from airplane.exceptions import InvalidEnvironmentException, RunPendingException
from airplane.output import append_output, set_output, write_named_output, write_output
from airplane.params import LabeledOption, ParamConfig
from airplane.runtime import execute, prompt
from airplane.runtime.standard import run  # Deprecated
from airplane.types import SQL, ConfigVar, File, LongText



from typing import Any, Dict, List, Optional

import backoff
import deprecation
import requests

from airplane._version import __version__
from airplane.api.client import api_client_from_env
from airplane.api.entities import PromptReviewers, Run, RunStatus, TaskReviewer
from airplane.exceptions import (
    TASK_MUST_BE_REQUESTED_ERROR_CODE,
    HTTPError,
    PromptPendingException,
    RequestPendingException,
    RequestRejectedException,
    RunPendingException,
    RunTerminationException,
)
from airplane.params import ParamTypes, SerializedParam


def execute(
    slug: str,
    param_values: Optional[Dict[str, ParamTypes]] = None,
    resources: Optional[Dict[str, Any]] = None,
) -> Run:
    """Standard executes an Airplane task, waits for execution, and returns run metadata.

    Args:
        slug: The slug of the task to run.
        param_values: Optional map of parameter slugs to values.
        resources: Optional map of resource aliases to ids.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the task cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
        ValueError: If the task is missing a form trigger.
        RequestRejectedException: If the request for the task is rejected.
    """

    client = api_client_from_env()
    try:
        run_id = client.execute_task(slug, param_values, resources)
    except HTTPError as err:
        if err.error_code != TASK_MUST_BE_REQUESTED_ERROR_CODE:
            if err.status_code >= 400 and err.status_code < 500:
                err.message = f'Failed to execute task "{slug}": {err.message}'
            raise
        task_reviewers_info = client.get_task_reviewers(slug)
        form_trigger = None
        if "task" in task_reviewers_info:
            for trigger in task_reviewers_info["task"].get("triggers", []):
                if trigger["kind"] == "form":
                    form_trigger = trigger

        if form_trigger is None:
            # pylint: disable=raise-missing-from
            raise ValueError("Missing form trigger for task, unable to create request")

        reviewers = task_reviewers_info.get("reviewers")
        if reviewers:
            reviewers = [
                TaskReviewer(user_id=r.get("userID"), group_id=r.get("groupID"))
                for r in reviewers
            ]
        trigger_request_id = client.create_task_request(
            trigger_id=form_trigger["triggerID"],
            param_values=param_values,
            reason="Automatically generated from parent run.",
            reviewers=reviewers,
        )

        trigger_request_info = __wait_for_request_completion(trigger_request_id)
        if trigger_request_info["status"] == "rejected":
            # pylint: disable=raise-missing-from
            raise RequestRejectedException(f"Request for task {slug} was rejected")

        run_id = ""
        if "triggerReceipt" in trigger_request_info:
            run_id = trigger_request_info["triggerReceipt"].get("taskRunID", "")

        if not run_id:
            # pylint: disable=raise-missing-from
            raise HTTPError(
                message="Unable to find run ID for completed request",
                status_code=err.status_code,
                error_code=err.error_code,
            )

    run_info = __wait_for_run_completion(run_id)
    use_zone = run_info.get("zoneID", None) is not None
    if use_zone:
        outputs = client.get_run_output_from_zone(run_id)
    else:
        outputs = client.get_run_output(run_id)

    # pylint: disable=redefined-outer-name
    run = Run(
        id=run_info["id"],
        task_id=run_info.get("taskID", None),
        param_values=run_info["paramValues"],
        status=RunStatus(run_info["status"]),
        output=outputs,
    )

    if run.status in {RunStatus.FAILED, RunStatus.CANCELLED}:
        raise RunTerminationException(run, slug)

    return run


@deprecation.deprecated(
    deprecated_in="0.3.2",
    current_version=__version__,
    details="Use execute(slug, param_values) instead.",
)
def run(
    task_id: str,
    parameters: Optional[Dict[str, Any]] = None,
    env: Optional[Dict[str, Any]] = None,
    constraints: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    """Creates an Airplane run, waits for execution, and returns its output and status.

    Args:
        task_id: The id of the task to run.
        parameters: Optional map of parameter slugs to values.
        env: Optional map of environment variables.
        constraints: Optional map of run constraints.

    Returns:
        The status and outputs of the run.

    Raises:
        HTTPError: If the run cannot be created or executed properly.
    """
    client = api_client_from_env()
    run_id = client.create_run(task_id, parameters, env, constraints)
    run_info = __wait_for_run_completion(run_id)
    use_zone = run_info.get("zoneID", None) is not None
    if use_zone:
        outputs = client.get_run_output_from_zone(run_id)
    else:
        outputs = client.get_run_output(run_id)

    return {"status": run_info["status"], "outputs": outputs}


@backoff.on_exception(
    lambda: backoff.expo(factor=0.1, max_value=5),
    (
        requests.exceptions.ConnectionError,
        requests.exceptions.Timeout,
        RunPendingException,
    ),
)
def __wait_for_run_completion(run_id: str) -> Dict[str, Any]:
    client = api_client_from_env()
    run_info = client.get_run(run_id)
    if run_info["status"] in ("NotStarted", "Queued", "Active"):
        raise RunPendingException()
    return run_info


@backoff.on_exception(
    lambda: backoff.expo(factor=0.1, max_value=5),
    (
        requests.exceptions.ConnectionError,
        requests.exceptions.Timeout,
        RequestPendingException,
    ),
)
def __wait_for_request_completion(trigger_request_id: str) -> Dict[str, Any]:
    client = api_client_from_env()
    trigger_request_info = client.get_trigger_request(trigger_request_id)
    if trigger_request_info["status"] == "pending":
        raise RequestPendingException()
    return trigger_request_info


def prompt_background(
    serialized_params: List[SerializedParam],
    *,
    reviewers: Optional[PromptReviewers] = None,
    confirm_text: Optional[str] = None,
    cancel_text: Optional[str] = None,
    description: Optional[str] = None,
) -> str:
    """Creates a prompt in the background, returning the prompt ID."""

    client = api_client_from_env()
    return client.create_prompt(
        parameters=serialized_params,
        reviewers=reviewers,
        confirm_text=confirm_text,
        cancel_text=cancel_text,
        description=description,
    )


@backoff.on_exception(
    lambda: backoff.expo(factor=0.1, max_value=5),
    (
        requests.exceptions.ConnectionError,
        requests.exceptions.Timeout,
        PromptPendingException,
    ),
)
def wait_for_prompt(prompt_id: str) -> Dict[str, Any]:
    """Waits until a prompt is submitted and returns the prompt values."""
    client = api_client_from_env()
    prompt_info = client.get_prompt(prompt_id)
    if not prompt_info["submittedAt"]:
        raise PromptPendingException()
    return prompt_info



from typing import Any, Dict, List, Optional, cast

from typing_extensions import TypedDict

from airplane.api.entities import BuiltInRun
from airplane.builtins import __convert_resource_alias_to_id
from airplane.runtime import __execute_internal


def find(
    mongodb_resource: str,
    collection: str,
    filter: Optional[Dict[str, Any]] = None,  # pylint: disable=redefined-builtin
    projection: Optional[Dict[str, Any]] = None,
    sort: Optional[Dict[str, Any]] = None,
    skip: Optional[int] = None,
    limit: Optional[int] = None,
) -> BuiltInRun[List[Dict[str, Any]]]:
    """Runs the find function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to search in.
        filter: The query predicate.
        projection: The projection specification that determines which fields to return.
        sort: The sort specification for the ordering of the results.
        skip: Number of documents to skip.
        limit: The maximum number of documents to return.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the find builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return cast(
        BuiltInRun[List[Dict[str, Any]]],
        __execute_internal(
            "airplane:mongodb_find",
            {
                "collection": collection,
                "filter": filter,
                "projection": projection,
                "sort": sort,
                "skip": skip,
                "limit": limit,
            },
            {"db": __convert_resource_alias_to_id(mongodb_resource)},
        ),
    )


def find_one(
    mongodb_resource: str,
    collection: str,
    filter: Optional[Dict[str, Any]] = None,  # pylint: disable=redefined-builtin
    projection: Optional[Dict[str, Any]] = None,
    sort: Optional[Dict[str, Any]] = None,
) -> BuiltInRun[Optional[Dict[str, Any]]]:
    """Runs the findOne function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to search.
        filter: The query predicate.
        projection: The projection specification that determines which fields to return.
        sort: The sort specification for the ordering of the results.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the findOne builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return cast(
        BuiltInRun[Optional[Dict[str, Any]]],
        __execute_internal(
            "airplane:mongodb_findOne",
            {
                "collection": collection,
                "filter": filter,
                "projection": projection,
                "sort": sort,
            },
            {"db": __convert_resource_alias_to_id(mongodb_resource)},
        ),
    )


def find_one_and_delete(
    mongodb_resource: str,
    collection: str,
    filter: Optional[Dict[str, Any]] = None,  # pylint: disable=redefined-builtin
    projection: Optional[Dict[str, Any]] = None,
    sort: Optional[Dict[str, Any]] = None,
) -> BuiltInRun[Optional[Dict[str, Any]]]:
    """Runs the findOneAndDelete function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to search in.
        filter: The query predicate.
        projection: The projection specification that determines which fields to return.
        sort: The sort specification for the ordering of the results.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the findOneAndDelete builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return cast(
        BuiltInRun[Optional[Dict[str, Any]]],
        __execute_internal(
            "airplane:mongodb_findOneAndDelete",
            {
                "collection": collection,
                "filter": filter,
                "projection": projection,
                "sort": sort,
            },
            {"db": __convert_resource_alias_to_id(mongodb_resource)},
        ),
    )


def find_one_and_update(
    mongodb_resource: str,
    collection: str,
    update: Dict[str, Any],
    filter: Optional[Dict[str, Any]] = None,  # pylint: disable=redefined-builtin
    projection: Optional[Dict[str, Any]] = None,
    sort: Optional[Dict[str, Any]] = None,
) -> BuiltInRun[Optional[Dict[str, Any]]]:
    """Runs the findOneAndUpdate function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to search in.
        update: The update document.
        filter: The query predicate.
        projection: The projection specification that determines which fields to return.
        sort: The sort specification for the ordering of the results.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the findOneAndUpdate builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return cast(
        BuiltInRun[Optional[Dict[str, Any]]],
        __execute_internal(
            "airplane:mongodb_findOneAndUpdate",
            {
                "collection": collection,
                "update": update,
                "filter": filter,
                "projection": projection,
                "sort": sort,
            },
            {"db": __convert_resource_alias_to_id(mongodb_resource)},
        ),
    )


def find_one_and_replace(
    mongodb_resource: str,
    collection: str,
    replacement: Dict[str, Any],
    filter: Optional[Dict[str, Any]] = None,  # pylint: disable=redefined-builtin
    projection: Optional[Dict[str, Any]] = None,
    sort: Optional[Dict[str, Any]] = None,
    upsert: Optional[bool] = None,
) -> BuiltInRun[Optional[Dict[str, Any]]]:
    """Runs the findOneAndReplace function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to search in.
        replacement: The replacement document.
        filter: The query predicate.
        projection: The projection specification that determines which fields to return.
        sort: The sort specification for the ordering of the results.
        upsert: Replaces the document or inserts the replacement if no document is found.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the findOneAndReplace builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return cast(
        BuiltInRun[Optional[Dict[str, Any]]],
        __execute_internal(
            "airplane:mongodb_findOneAndReplace",
            {
                "collection": collection,
                "replacement": replacement,
                "filter": filter,
                "projection": projection,
                "sort": sort,
                "upsert": upsert,
            },
            {"db": __convert_resource_alias_to_id(mongodb_resource)},
        ),
    )


class InsertOneOutput(TypedDict):
    """The output of the mongodb.insert_one builtin."""

    InsertedID: str


def insert_one(
    mongodb_resource: str,
    collection: str,
    document: Dict[str, Any],
) -> BuiltInRun[InsertOneOutput]:
    """Runs the insertOne function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to insert in.
        document: The document to insert.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the insertOne builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return cast(
        BuiltInRun[InsertOneOutput],
        __execute_internal(
            "airplane:mongodb_insertOne",
            {
                "collection": collection,
                "document": document,
            },
            {"db": __convert_resource_alias_to_id(mongodb_resource)},
        ),
    )


class InsertManyOutput(TypedDict):
    """The output of the mongodb.insert_many builtin."""

    InsertedIDs: List[str]


def insert_many(
    mongodb_resource: str,
    collection: str,
    documents: List[Dict[str, Any]],
) -> BuiltInRun[InsertManyOutput]:
    """Runs the insertMany function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to insert in.
        documents: The documents to insert.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the insertMany builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return cast(
        BuiltInRun[InsertManyOutput],
        __execute_internal(
            "airplane:mongodb_insertMany",
            {
                "collection": collection,
                "documents": documents,
            },
            {"db": __convert_resource_alias_to_id(mongodb_resource)},
        ),
    )


class UpdateOutput(TypedDict):
    """The output of the mongodb.update_one and mongodb.udpate_many builtins."""

    MatchedCount: int
    ModifiedCount: int
    UpsertedCount: int
    UpsertedID: Optional[str]


def update_one(
    mongodb_resource: str,
    collection: str,
    update: Dict[str, Any],
    filter: Optional[Dict[str, Any]] = None,  # pylint: disable=redefined-builtin
    upsert: Optional[bool] = None,
) -> BuiltInRun[UpdateOutput]:
    """Runs the updateOne function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to update in.
        update: The update document.
        filter: The query predicate.
        upsert: Updates the document or inserts the update if no document is found.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the updateOne builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return cast(
        BuiltInRun[UpdateOutput],
        __execute_internal(
            "airplane:mongodb_updateOne",
            {
                "collection": collection,
                "update": update,
                "filter": filter,
                "upsert": upsert,
            },
            {"db": __convert_resource_alias_to_id(mongodb_resource)},
        ),
    )


def update_many(
    mongodb_resource: str,
    collection: str,
    update: Dict[str, Any],
    filter: Optional[Dict[str, Any]] = None,  # pylint: disable=redefined-builtin
    upsert: Optional[bool] = None,
) -> BuiltInRun[UpdateOutput]:
    """Runs the updateMany function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to update in.
        update: The update document.
        filter: The query predicate.
        upsert: Updates the documents or inserts the update if no document is found.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the updateMany builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return cast(
        BuiltInRun[UpdateOutput],
        __execute_internal(
            "airplane:mongodb_updateMany",
            {
                "collection": collection,
                "update": update,
                "filter": filter,
                "upsert": upsert,
            },
            {"db": __convert_resource_alias_to_id(mongodb_resource)},
        ),
    )


class DeleteOutput(TypedDict):
    """The output of the mongodb.delete_one and mongodb.delete_many builtins."""

    DeletedCount: int


def delete_one(
    mongodb_resource: str,
    collection: str,
    filter: Dict[str, Any],  # pylint: disable=redefined-builtin
) -> BuiltInRun[DeleteOutput]:
    """Runs the deleteOne function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to delete in.
        filter: The query predicate.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the deleteOne builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return cast(
        BuiltInRun[DeleteOutput],
        __execute_internal(
            "airplane:mongodb_deleteOne",
            {
                "collection": collection,
                "filter": filter,
            },
            {"db": __convert_resource_alias_to_id(mongodb_resource)},
        ),
    )


def delete_many(
    mongodb_resource: str,
    collection: str,
    filter: Dict[str, Any],  # pylint: disable=redefined-builtin
) -> BuiltInRun[DeleteOutput]:
    """Runs the deleteMany function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to delete in.
        filter: The query predicate.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the deleteMany builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return cast(
        BuiltInRun[DeleteOutput],
        __execute_internal(
            "airplane:mongodb_deleteMany",
            {
                "collection": collection,
                "filter": filter,
            },
            {"db": __convert_resource_alias_to_id(mongodb_resource)},
        ),
    )


def aggregate(
    mongodb_resource: str,
    collection: str,
    pipeline: List[Dict[str, Any]],
) -> BuiltInRun[List[Dict[str, Any]]]:
    """Runs the aggregate function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to aggregate in.
        pipeline: The sequence of data aggregation operations.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the aggregate builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return cast(
        BuiltInRun[List[Dict[str, Any]]],
        __execute_internal(
            "airplane:mongodb_aggregate",
            {
                "collection": collection,
                "pipeline": pipeline,
            },
            {"db": __convert_resource_alias_to_id(mongodb_resource)},
        ),
    )


def count_documents(
    mongodb_resource: str,
    collection: str,
    filter: Dict[str, Any],  # pylint: disable=redefined-builtin
) -> BuiltInRun[float]:
    """Runs the countDocuments function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to count in.
        filter: The query predicate.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the countDocuments builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return cast(
        BuiltInRun[float],
        __execute_internal(
            "airplane:mongodb_countDocuments",
            {
                "collection": collection,
                "filter": filter,
            },
            {"db": __convert_resource_alias_to_id(mongodb_resource)},
        ),
    )


def distinct(
    mongodb_resource: str,
    collection: str,
    field: str,
    filter: Dict[str, Any],  # pylint: disable=redefined-builtin
) -> BuiltInRun[List[Any]]:
    """Runs the distinct function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to search.
        field: The field for which to return distinct values.
        filter: The query predicate.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the distinct builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return cast(
        BuiltInRun[List[Any]],
        __execute_internal(
            "airplane:mongodb_distinct",
            {
                "collection": collection,
                "field": field,
                "filter": filter,
            },
            {"db": __convert_resource_alias_to_id(mongodb_resource)},
        ),
    )



import inflection
from slugify import slugify  # type: ignore


def make_slug(string: str) -> str:
    """Turns a string into a slug"""
    for target, replacement in [
        ("‒", "_"),  # figure dash
        ("–", "_"),  # en dash
        ("—", "_"),  # em dash
        ("―", "_"),  # horizontal bar
        ("&", "_and_"),
        ("@", "_at_"),
        ("%", "_percent_"),
    ]:
        string = string.replace(target, replacement)
    return slugify(inflection.underscore(string)).replace("-", "_")[:50]



import dataclasses
import datetime
from typing import Any, Generic, List, Optional, Sequence, Tuple, Type, TypeVar, Union

from typing_extensions import Annotated, Literal, get_args, get_origin

from airplane.exceptions import InvalidAnnotationException
from airplane.types import SQL, ConfigVar, File, LongText

SERIALIZED_DATE_FORMAT = "%Y-%m-%d"
SERIALIZED_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
SERIALIZED_DATETIME_MILLISECONDS_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


ParamType = Literal[
    "shorttext",
    "longtext",
    "sql",
    "boolean",
    "upload",
    "integer",
    "float",
    "date",
    "datetime",
    "configvar",
]

ParamTypes = Union[
    str,
    LongText,
    SQL,
    bool,
    File,
    int,
    float,
    datetime.date,
    datetime.datetime,
    ConfigVar,
]

PromptParamTypes = Union[
    Type[str],
    Type[LongText],
    Type[SQL],
    Type[bool],
    Type[File],
    Type[int],
    Type[float],
    Type[datetime.date],
    Type[datetime.datetime],
    Type[ConfigVar],
    # To support Annotated types
    object,
]

SerializedParamType = Union[
    Literal["string"],
    Literal["boolean"],
    Literal["upload"],
    Literal["integer"],
    Literal["float"],
    Literal["date"],
    Literal["datetime"],
    Literal["configvar"],
]


SerializedParamComponent = Union[
    Literal["editor-sql"],
    Literal["textarea"],
]

SerializedParamValue = Union[str, bool, int, float, None]


@dataclasses.dataclass(frozen=True)
class Constraints:
    """Parameter constraints."""

    optional: bool
    regex: Optional[str] = None
    options: Optional[Any] = None


@dataclasses.dataclass(frozen=True)
class SerializedParam:
    """Serialized parameter."""

    slug: str
    name: str
    type: SerializedParamType
    constraints: Constraints
    desc: Optional[str] = None
    component: Optional[SerializedParamComponent] = None
    default: Optional[SerializedParamValue] = None


DefaultParamT = TypeVar(
    "DefaultParamT",
    bound=ParamTypes,
)


@dataclasses.dataclass(frozen=True)
class LabeledOption(Generic[DefaultParamT]):
    """Parmeter select option with a label."""

    label: str
    value: DefaultParamT


OptionsT = Sequence[Union[DefaultParamT, LabeledOption[DefaultParamT]]]

AllOptions = Union[
    OptionsT[str],
    OptionsT[int],
    OptionsT[float],
    OptionsT[datetime.date],
    OptionsT[datetime.datetime],
    OptionsT[ConfigVar],
]


@dataclasses.dataclass(frozen=True)
class ParamConfig:
    """Task parameter configuration.

    Attributes:
        slug:
            Human-friendly identifier used to reference this parameter. Parameter slugs
            must be unique within an individual task / workflow. Defaults to the function
            argument's name.
        name:
            Parameter name displayed on the Airplane app. Defaults to the function
            argument's name in sentence case.
        description:
            Parameter description displayed on the Airplane app. If not provided, the description
            will be pulled from the docstring of the decorated function.
        options:
            Option constraint for the parameter. Select options allow users to specify exactly
            which values are allowed for a given parameter. Options may specify a label which
            is shown to the user on the Airplane app instead of the value.
        regex:
            Regex contraint for the parameter, only valid for string arguments.
        default:
            Default value for the parameter. This field should only be used for prompt
            parameters, not task parameters. For task parameters, use the Python native
            default function parameter syntax, .e.g. `def my_task(my_param: str = "default")`.
    """

    slug: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    options: Optional[AllOptions] = None
    regex: Optional[str] = None
    default: Optional[ParamTypes] = None


def to_airplane_type(
    param_name: str,
    type_hint: Any,
    func_name: Optional[str] = None,
) -> ParamType:
    """Converts a Python type hint to an Airplane type."""
    if type_hint == str:
        return "shorttext"
    if type_hint == LongText:
        return "longtext"
    if type_hint == SQL:
        return "sql"
    if type_hint == bool:
        return "boolean"
    if type_hint == File:
        return "upload"
    if type_hint == int:
        return "integer"
    if type_hint == float:
        return "float"
    if type_hint == datetime.date:
        return "date"
    if type_hint == datetime.datetime:
        return "datetime"
    if type_hint == ConfigVar:
        return "configvar"

    raise InvalidAnnotationException(
        prefix=f"Invalid type annotation `{type_hint}`",
        func_name=func_name,
        param_name=param_name,
    )


def to_serialized_airplane_type(
    param_name: str,
    type_hint: Any,
    func_name: Optional[str] = None,
) -> Tuple[SerializedParamType, Optional[SerializedParamComponent]]:
    """Converts a Python type hint to a serialized Airplane type."""
    if type_hint == str:
        return "string", None
    if type_hint == LongText:
        return "string", "textarea"
    if type_hint == SQL:
        return "string", "editor-sql"
    if type_hint == bool:
        return "boolean", None
    if type_hint == File:
        return "upload", None
    if type_hint == int:
        return "integer", None
    if type_hint == float:
        return "float", None
    if type_hint == datetime.date:
        return "date", None
    if type_hint == datetime.datetime:
        return "datetime", None
    if type_hint == ConfigVar:
        return "configvar", None

    raise InvalidAnnotationException(
        prefix=f"Invalid type annotation `{type_hint}`",
        func_name=func_name,
        param_name=param_name,
    )


def resolve_type(
    param_name: str,
    type_hint: Any,
    func_name: Optional[str] = None,
) -> Tuple[Any, bool, Optional[ParamConfig]]:
    """Parses a parameter's type hint to extract its underlying type,
    whether it's optional, and a ParamConfig if provided."""

    origin_type = get_origin(type_hint)
    if origin_type == Union:
        type_args = get_args(type_hint)
        if len(type_args) != 2 or type_args[1] is not type(None):
            raise InvalidAnnotationException(
                prefix=f"Unsupported Union type `{type_hint}`",
                func_name=func_name,
                param_name=param_name,
            )
        underlying_type, _, param_config = resolve_type(
            param_name, type_args[0], func_name
        )
        return underlying_type, True, param_config

    if origin_type == Annotated:
        type_args = get_args(type_hint)
        param_configs = [t for t in type_args if isinstance(t, ParamConfig)]
        # Don't support multiple parameter configs within the same annotation.
        if len(param_configs) > 1:
            raise InvalidAnnotationException(
                prefix=f"Found multiple ParamConfig annotations `{type_hint}`",
                func_name=func_name,
                param_name=param_name,
            )
        param_config = param_configs[0] if param_configs else None
        underlying_type, is_optional, _ = resolve_type(
            param_name, type_args[0], func_name
        )
        return underlying_type, is_optional, param_config
    return type_hint, False, None


ParamDefTypes = Union[str, int, float]

# ParamDefs have a subset of types that are built in and serializable.
ParamDefOptions = Union[
    List[LabeledOption[str]],
    List[LabeledOption[int]],
    List[LabeledOption[float]],
]


def serialize_param(
    val: ParamTypes,
) -> ParamDefTypes:
    """Transforms a general parameter into a format supported by parameter definition"""
    if isinstance(val, datetime.datetime):
        return val.strftime(SERIALIZED_DATETIME_FORMAT)
    if isinstance(val, datetime.date):
        return val.strftime(SERIALIZED_DATE_FORMAT)
    if isinstance(val, ConfigVar):
        return val.name
    if isinstance(val, File):
        return val.id
    return val


def make_options(param_config: ParamConfig) -> Optional[ParamDefOptions]:
    """Builds a list of options for a parameter definition"""
    options: Optional[ParamDefOptions]
    if param_config.options is None:
        options = None
    else:
        options = []  # type: ignore
        for option in param_config.options:
            if isinstance(option, LabeledOption):
                labeled_option = LabeledOption(
                    label=option.label,
                    value=serialize_param(option.value),
                )
            else:
                value = serialize_param(option)
                labeled_option = LabeledOption(
                    label=str(value),
                    value=value,
                )
            options.append(labeled_option)  # type: ignore
    return options
